==> spirit_hub-1.0.0.5.tar.gz: spirit_hub-1.0.0.5/app/overrides/tagset.rb <==
# frozen_string_literal: true

Tagset.class_eval do # rubocop:disable Metrics/BlockLength
  include Commentable
  field :operation_title, type: String
  field :phase_title, type: String
  field :type, type: String, default: 'directory'
  field :spirit_hubdoc, type: Mongoid::Boolean, default: false
  field :file_ext, type: String
  field :doc_file_size, type: Integer
  field :last_current_user_id, type: String
  field :last_current_user_name, type: String
  field :current_state, type: String
  field :start_date, type: Date
  field :end_date, type: Date
  field :content_completeness, type: Integer
  field :doc_ids, type: Array, default: []
  field :parent_tagset_ids, type: Array, default: []
  field :informed_main, type: Array, default: []
  field :informed_cc, type: Array, default: []
  field :informed_bcc, type: Array, default: []
  field :max_duplicates, type: String
  field :existing_duplicates, type: Array, default: []
  field :notifiable_ids, type: Array, default: []

  belongs_to :operation, index: true, optional: true, class_name: 'SpiritHub::Operation'
  belongs_to :phase, index: true, optional: true, class_name: 'SpiritHub::Phase'

  has_many :journal_entries, class_name: 'SpiritHub::JournalEntry', dependent: :destroy

  before_create :update_directory_completeness
  before_save :spirit_hub_before_save
  after_save :update_children_parent_ids, :update_phase_percent
  after_destroy :update_completeness, :add_delete_log

  index({ type: 1 }, { background: true })
  index({ last_current_user_id: 1 }, { background: true })
  index({ current_state: 1 }, { background: true })

  index({ spirit_hubdoc: 1 }, { background: true }) # spirit_hubdoc index
  index({ spirit_hubdoc: 1, type: 1, all_can_access: 1 }, { background: true }) # access index
  index({ spirit_hubdoc: 1, type: 1, percent: 1, accounted: 1 }, { background: true }) # todos index
  index({ spirit_hubdoc: 1, type: 1, responsible: 1, parent_tagset_ids: 1 }, { background: true }) # to_classify index

  scope :to_classify, -> { where(type: 'document', parent_tagset_ids: { '$size': 0 }, doc_ids: { "$ne": [] }) }
  scope :to_be_processed, ->(user) { where(type: 'document', accounted: user.id, percent: { '$ne': 100 }) }

  def self.create_tagset_docs_from_files(files, user, **opts)
    files.map { |file| create_tagset_doc_from_file(file, user, **opts) }
  end

  def self.create_tagset_doc_from_file(files, user, **opts)
    doc_ids = []

    Array(files).each do |file|
      doc_ids << Doc.create_from_file!(file, user).id
    end

    title = opts.fetch(:title, nil)
    if title.blank?
      file = Array(files).first
      title = file.original_filename if file.respond_to?(:original_filename)
      title = File.basename(file) if title.blank?
    end

    tagset_doc = Tagset.create!(
      doc_ids:,
      responsible: user.id,
      type: 'document',
      spirit_hubdoc: true,
      title:,
      cycle: SpiritHub::HubdocCycle.default_cycle
    )

    opts.fetch(:parents, []).each { |parent| tagset_doc.add_tagset_doc_to_tagset(parent, user, persist: false) }

    SpiritHub::JournalEntry.create_action_log('DOC_CREATED', tagset: tagset_doc, user:)
    SpiritHub::HubdocCycle.change_state!(tagset_doc, opts[:state], user) if opts[:state].present?

    tagset_doc
  end

  def doc?
    type == 'document'
  end

  def one_doc?
    doc? && docs.pluck(:id).to_a.uniq.compact.one?
  end

  def multiple_doc?
    doc? && docs.pluck(:id).to_a.uniq.compact.length > 1
  end

  def directory?
    type == 'directory'
  end

  def completed?
    return content_completeness == 100 if directory?

    true # if not a directory, it doesn't actually matter
  end

  def all_docs
    Doc.where(:id.in => doc_ids)
  end

  def docs
    Doc.unversioned.where(:id.in => doc_ids)
  end

  def doc
    docs.first if one_doc?
  end

  def doc_id
    doc_ids.first if one_doc?
  end

  def find_docs_or_children_docs
    return docs if doc?

    all_children_tagsets.flat_map(&:docs)
  end

  def children_tagsets
    Tagset.where(:id.in => Array(tagset_ids))
  end

  def all_children_tagsets
    children_tagsets.flat_map { |child| [child] + child.all_children_tagsets }
  end

  def parents
    Tagset.where(:id.in => Array(parent_tagset_ids))
  end

  def parent
    Tagset.find_by(id: parent_tagset_ids.to_a.first)
  end

  def parent_directories
    parents.where(type: 'directory')
  end

  def completeness
    SpiritHub::Completeness.where(tagset_id: id)
  end

  def uncompleted_completeness
    SpiritHub::Completeness.where(tagset_id: id, completed: false)
  end

  def current_user
    User.find_by(id: current_user_id)
  end

  def update_docs_fields
    docs_count = docs.to_a.size
    self.file_ext = case docs_count
                    when 0
                      nil
                    when 1
                      docs.last.file_ext
                    else
                      'multiple'
                    end
    self.doc_file_size = docs_count.zero? ? nil : docs.sum(:file_size)
  end

  def update_operation_title
    self.operation_title = operation&.title
  end

  def update_phase_title
    return unless directory?

    self.phase_title = phase&.title
  end

  def remove_doc_from_tagset_doc(doc, user)
    doc.current_user_id = user.id
    doc.destroy
    doc_ids.delete(doc.id)
    save
  end

  def add_doc_to_tagset_doc(files, user)
    new_doc_ids = []

    Array(files).each do |file|
      new_doc_ids << Doc.create_from_file!(file, user).id
    end

    self.doc_ids = doc_ids.to_a + new_doc_ids
    save
  end

  def add_tagset_doc_to_object_by_id(object_id, user)
    if object_id.to_s.starts_with?('completeness')
      add_tagset_doc_to_completeness(object_id.gsub('completeness_', ''), user)
    elsif object_id.present?
      add_tagset_doc_to_tagset(object_id, user)
    end
  end

  def add_tagset_doc_to_completeness(completeness_or_id, user, persist: true)
    completeness = completeness_or_id if completeness_or_id.is_a?(SpiritHub::Completeness)
    completeness ||= SpiritHub::Completeness.find_by(id: completeness_or_id)

    SpiritHub::TagsetPurger.clean_linked_models(self, user) if persist

    completeness.update(object_ids: completeness.object_ids.to_a + [id])
    SpiritHub::JournalEntry.create_action_log('DOC_ADDED_TO_COMPLETENESS', tagset: self, user:, title: completeness.title)

    self.title = completeness&.title
    add_tagset_doc_to_tagset(completeness.tagset, user, persist: false)

    save if persist
  end

  def add_tagset_doc_to_tagset(tagset_or_id, user, persist: true)
    tagset = tagset_or_id if tagset_or_id.is_a?(Tagset)
    tagset ||= Tagset.find_by(id: tagset_or_id)

    SpiritHub::TagsetPurger.clean_linked_models(self, user) if persist

    tagset.update(tagset_ids: tagset.tagset_ids.to_a + [id.to_s])
    SpiritHub::JournalEntry.classify_action_log('DOC_CLASSIFIED', tagset: self, user:, directory: tagset.title)

    self.operation_id = tagset.operation_id
    SpiritHub::HubdocCycle.change_state(self, :classify, user)
    return unless persist

    save
    SpiritHub::TagsetNotifier.call(reload, user)
  end

  def parent_tagset_title
    Tagset.find_by(id: parent_tagset_ids.to_a.first)&.title
  end

  # TODO : Move to core
  def all_can_access?(user)
    (all_can_access.to_a & user.all_gids_and_user).any?
  end

  def writers?(user)
    (writers.to_a & user.all_gids_and_user).any?
  end

  def readers?(user)
    (readers.to_a & user.all_gids_and_user).any?
  end

  def accounted?(user)
    (accounted.to_a & user.all_gids_and_user).any?
  end

  def responsible?(user)
    responsible&.to_s == user.id.to_s
  end

  private

  def add_delete_log
    return unless spirit_hubdoc

    log_code = doc? ? 'DOC_DELETED' : 'DIRECTORY_DELETED'
    SpiritHub::JournalEntry.delete_action_log(log_code, tagset: self, user: current_user, title:)
  end

  def spirit_hub_before_save
    update_operation_title if will_save_change_to_attribute?(:operation_id)
    update_phase_title if will_save_change_to_attribute?(:phase_id)
    update_docs_fields if will_save_change_to_attribute?(:doc_ids)
    update_informed_fields if will_save_change_to_any_of_the_informed_attributes?
    update_directory_completeness if will_save_change_to_attribute?(:tagset_ids)
    update_directory_completeness if will_save_change_to_attribute?(:current_state) # TODO : remove when we have a proper state management
  end

  def update_completeness
    SpiritHub::Completeness.where(object_ids: id).each do |completeness|
      completeness.update(object_ids: completeness.object_ids - [id])
    end
  end

  def will_save_change_to_any_of_the_informed_attributes?
    %i[informed informed_main informed_cc informed_bcc].any? { |attr| will_save_change_to_attribute?(attr) }
  end

  def update_informed_fields
    # NOTE: We now have a couple of informed-related columns. Basically we have to differentiate which
    # are the main, cc anc bcc informed users. We also want to keep the backwards compatibility with the
    # original :informed field. For this reason, if the original field is changed, we assum all are the main
    # ones and reset the other ones. Otherwise, we set the original informed field based on the other ones.
    if will_save_change_to_informed?
      self.informed_main = Array(informed)
      self.informed_cc = []
      self.informed_bcc = []
    elsif will_save_change_to_informed_main? || will_save_change_to_informed_cc? || will_save_change_to_informed_bcc?
      self.informed = Array(informed_main) + Array(informed_cc) + Array(informed_bcc)
    end
  end

  def update_children_parent_ids
    old_tagset_ids = previous_changes['tagset_ids'].to_a.first || []
    new_tagset_ids = previous_changes['tagset_ids'].to_a.last || []

    to_remove = old_tagset_ids - new_tagset_ids
    to_add = new_tagset_ids - old_tagset_ids

    to_add.each do |child_id|
      Tagset.where(id: child_id).push(parent_tagset_ids: id)
    end

    to_remove.each do |child_id|
      Tagset.where(id: child_id).pull(parent_tagset_ids: id)
    end
  end

  def update_phase_percent
    return unless directory?
    return unless saved_change_to_attribute?(:content_completeness) || saved_change_to_attribute?(:phase_id)

    phase&.update_percentage

    return unless (old_phase = SpiritHub::Phase.find_by(id: previous_changes['phase_id']&.first))

    old_phase.update_percentage
  end

  def update_directory_completeness
    return unless directory?

    SpiritHub::DirectoryCompletenessUpdater.call(self, skip_save: true)
  end
end

==> spirit_hub-1.0.0.15.tar.gz: spirit_hub-1.0.0.15/app/overrides/tagset.rb <==
# frozen_string_literal: true

Tagset.class_eval do # rubocop:disable Metrics/BlockLength
  include Commentable
  field :operation_title, type: String
  field :phase_title, type: String
  field :type, type: String, default: 'directory'
  field :spirit_hubdoc, type: Mongoid::Boolean, default: false
  field :file_ext, type: String
  field :doc_file_size, type: Integer
  field :last_current_user_id, type: String
  field :last_current_user_name, type: String
  field :current_state, type: String
  field :start_date, type: Date
  field :end_date, type: Date
  field :content_completeness, type: Integer
  field :doc_ids, type: Array, default: []
  field :parent_tagset_ids, type: Array, default: []
  field :informed_main, type: Array, default: []
  field :informed_cc, type: Array, default: []
  field :informed_bcc, type: Array, default: []
  field :max_duplicates, type: String
  field :existing_duplicates, type: Array, default: []
  field :notifiable_ids, type: Array, default: []

  belongs_to :operation, index: true, optional: true, class_name: 'SpiritHub::Operation'
  belongs_to :phase, index: true, optional: true, class_name: 'SpiritHub::Phase'

  has_many :journal_entries, class_name: 'SpiritHub::JournalEntry', dependent: :destroy

  before_create :update_directory_completeness
  before_save :spirit_hub_before_save
  after_save :update_children_parent_ids, :update_phase_percent
  after_destroy :update_completeness, :add_delete_log

  index({ type: 1 }, { background: true })
  index({ last_current_user_id: 1 }, { background: true })
  index({ current_state: 1 }, { background: true })

  index({ spirit_hubdoc: 1 }, { background: true }) # spirit_hubdoc index
  index({ spirit_hubdoc: 1, type: 1, all_can_access: 1 }, { background: true }) # access index
  index({ spirit_hubdoc: 1, type: 1, percent: 1, accounted: 1 }, { background: true }) # todos index
  index({ spirit_hubdoc: 1, type: 1, responsible: 1, parent_tagset_ids: 1 }, { background: true }) # to_classify index

  scope :to_classify, -> { where(type: 'document', parent_tagset_ids: { '$size': 0 }, doc_ids: { "$ne": [] }) }
  scope :to_be_processed, ->(user) { where(type: 'document', accounted: user.id, percent: { '$ne': 100 }) }

  def self.create_tagset_docs_from_files(files, user, **opts)
    files.map { |file| create_tagset_doc_from_file(file, user, **opts) }
  end

  def self.create_tagset_doc_from_file(files, user, **opts)
    doc_ids = []

    Array(files).each do |file|
      doc_ids << Doc.create_from_file!(file, user).id
    end

    title = opts.fetch(:title, nil)
    if title.blank?
      file = Array(files).first
      title = file.original_filename if file.respond_to?(:original_filename)
      title = File.basename(file) if title.blank?
    end

    tagset_doc = Tagset.create!(
      doc_ids:,
      responsible: user.id,
      type: 'document',
      spirit_hubdoc: true,
      title:,
      cycle: SpiritHub::HubdocCycle.default_cycle
    )

    opts.fetch(:parents, []).each { |parent| tagset_doc.add_tagset_doc_to_tagset(parent, user, persist: false) }

    SpiritHub::JournalEntry.create_action_log('DOC_CREATED', tagset: tagset_doc, user:)
    SpiritHub::HubdocCycle.change_state!(tagset_doc, opts[:state], user) if opts[:state].present?

    tagset_doc
  end

  def doc?
    type == 'document'
  end

  def one_doc?
    doc? && docs.pluck(:id).to_a.uniq.compact.one?
  end

  def multiple_doc?
    doc? && docs.pluck(:id).to_a.uniq.compact.length > 1
  end

  def directory?
    type == 'directory'
  end

  def completed?
    return content_completeness == 100 if directory?

    true # if not a directory, it doesn't actually matter
  end

  def all_docs
    Doc.where(:id.in => doc_ids)
  end

  def docs
    Doc.unversioned.where(:id.in => doc_ids)
  end

  def doc
    docs.first if one_doc?
  end

  def doc_id
    doc_ids.first if one_doc?
  end

  def find_docs_or_children_docs
    return docs if doc?

    all_children_tagsets.flat_map(&:docs)
  end

  def children_tagsets
    Tagset.where(:id.in => Array(tagset_ids))
  end

  def all_children_tagsets
    children_tagsets.flat_map { |child| [child] + child.all_children_tagsets }
  end

  def parents
    Tagset.where(:id.in => Array(parent_tagset_ids))
  end

  def parent
    Tagset.find_by(id: parent_tagset_ids.to_a.first)
  end

  def parent_directories
    parents.where(type: 'directory')
  end

  def completeness
    SpiritHub::Completeness.where(tagset_id: id)
  end

  def uncompleted_completeness
    SpiritHub::Completeness.where(tagset_id: id, completed: false)
  end

  def current_user
    User.find_by(id: current_user_id)
  end

  def update_docs_fields
    docs_count = docs.to_a.size
    self.file_ext = case docs_count
                    when 0
                      nil
                    when 1
                      docs.last.file_ext
                    else
                      'multiple'
                    end
    self.doc_file_size = docs_count.zero? ? nil : docs.sum(:file_size)
  end

  def update_operation_title
    self.operation_title = operation&.title
  end

  def update_phase_title
    return unless directory?

    self.phase_title = phase&.title
  end

  def remove_doc_from_tagset_doc(doc, user)
    doc.current_user_id = user.id
    doc.destroy
    doc_ids.delete(doc.id)
    save
  end

  def add_doc_to_tagset_doc(files, user)
    new_doc_ids = []

    Array(files).each do |file|
      new_doc_ids << Doc.create_from_file!(file, user).id
    end

    self.doc_ids = doc_ids.to_a + new_doc_ids
    save
  end

  def add_tagset_doc_to_object_by_id(object_id, user, options = {})
    if object_id.to_s.start_with?('completeness')
      clean_id = object_id.sub(/^completeness_/, '')
      add_tagset_doc_to_completeness(clean_id, user, options)
    else
      add_tagset_doc_to_tagset(object_id, user, options)
    end
  end

  def add_tagset_doc_to_completeness(completeness_or_id, user, options = {})
    completeness = completeness_or_id.is_a?(SpiritHub::Completeness) ? completeness_or_id : SpiritHub::Completeness.find_by(id: completeness_or_id)
    return unless completeness

    persist = options.fetch(:persist, true)
    SpiritHub::TagsetPurger.clean_linked_models(self, user) if persist

    completeness.update!(object_ids: completeness.object_ids.to_a + [id])
    SpiritHub::JournalEntry.create_action_log('DOC_ADDED_TO_COMPLETENESS', tagset: self, user:, title: completeness.title)

    self.title = completeness.title if completeness.title
    add_tagset_doc_to_tagset(completeness.tagset, user, options.merge(persist: false)) if completeness.tagset

    save if persist
  end

  def add_tagset_doc_to_tagset(tagset_or_id, user, options = {})
    tagset = tagset_or_id.is_a?(Tagset) ? tagset_or_id : Tagset.find_by(id: tagset_or_id)
    return unless tagset

    persist = options.fetch(:persist, true)
    SpiritHub::TagsetPurger.clean_linked_models(self, user) if persist

    updated_ids = tagset.tagset_ids.to_a << id.to_s
    tagset.update(tagset_ids: updated_ids)
    SpiritHub::JournalEntry.classify_action_log('DOC_CLASSIFIED', tagset: self, user:, directory: tagset.title)

    self.operation_id = tagset.operation_id
    self.readers = tagset.readers.to_a
    self.writers = tagset.writers.to_a

    SpiritHub::HubdocCycle.change_state(self, :classify, user)

    return unless persist

    save!
    SpiritHub::TagsetNotifier.call(reload, user)
  end

  def parent_tagset_title
    Tagset.find_by(id: parent_tagset_ids.to_a.first)&.title
  end

  # TODO : Move to core
  def all_can_access?(user)
    (all_can_access.to_a & user.all_gids_and_user).any?
  end

  def writers?(user)
    (writers.to_a & user.all_gids_and_user).any?
  end

  def readers?(user)
    (readers.to_a & user.all_gids_and_user).any?
  end

  def accounted?(user)
    (accounted.to_a & user.all_gids_and_user).any?
  end

  def responsible?(user)
    responsible&.to_s == user.id.to_s
  end

  private

  def add_delete_log
    return unless spirit_hubdoc

    log_code = doc? ? 'DOC_DELETED' : 'DIRECTORY_DELETED'
    SpiritHub::JournalEntry.delete_action_log(log_code, tagset: self, user: current_user, title:)
  end

  def spirit_hub_before_save
    update_operation_title if will_save_change_to_attribute?(:operation_id)
    update_phase_title if will_save_change_to_attribute?(:phase_id)
    update_docs_fields if will_save_change_to_attribute?(:doc_ids)
    update_informed_fields if will_save_change_to_any_of_the_informed_attributes?
    update_directory_completeness if will_save_change_to_attribute?(:tagset_ids)
    update_directory_completeness if will_save_change_to_attribute?(:current_state) # TODO : remove when we have a proper state management
  end

  def update_completeness
    SpiritHub::Completeness.where(object_ids: id).each do |completeness|
      completeness.update(object_ids: completeness.object_ids - [id])
    end
  end

  def will_save_change_to_any_of_the_informed_attributes?
    %i[informed informed_main informed_cc informed_bcc].any? { |attr| will_save_change_to_attribute?(attr) }
  end

  def update_informed_fields
    # NOTE: We now have a couple of informed-related columns. Basically we have to differentiate which
    # are the main, cc anc bcc informed users. We also want to keep the backwards compatibility with the
    # original :informed field. For this reason, if the original field is changed, we assum all are the main
    # ones and reset the other ones. Otherwise, we set the original informed field based on the other ones.
    if will_save_change_to_informed?
      self.informed_main = Array(informed)
      self.informed_cc = []
      self.informed_bcc = []
    elsif will_save_change_to_informed_main? || will_save_change_to_informed_cc? || will_save_change_to_informed_bcc?
      self.informed = Array(informed_main) + Array(informed_cc) + Array(informed_bcc)
    end
  end

  def update_children_parent_ids
    old_tagset_ids = previous_changes['tagset_ids'].to_a.first || []
    new_tagset_ids = previous_changes['tagset_ids'].to_a.last || []

    to_remove = old_tagset_ids - new_tagset_ids
    to_add = new_tagset_ids - old_tagset_ids

    to_add.each do |child_id|
      Tagset.where(id: child_id).push(parent_tagset_ids: id)
    end

    to_remove.each do |child_id|
      Tagset.where(id: child_id).pull(parent_tagset_ids: id)
    end
  end

  def update_phase_percent
    return unless directory?
    return unless saved_change_to_attribute?(:content_completeness) || saved_change_to_attribute?(:phase_id)

    phase&.update_percentage

    return unless (old_phase = SpiritHub::Phase.find_by(id: previous_changes['phase_id']&.first))

    old_phase.update_percentage
  end

  def update_directory_completeness
    return unless directory?

    SpiritHub::DirectoryCompletenessUpdater.call(self, skip_save: true)
  end
end
